@@ -237,15 +237,6 @@ static ParseCallbackTableItem callback_table[] =
 static StringInfoData copyString;
 static bool copyStringInitDone = false;
 
-/*
- * Temporary buffer for storing decompressed data.
- *
- * 64K should be enough in most cases. If it's not user can manually change
- * this limit. Unfortunately there is no way to know how much memory user
- * is willing to allocate.
- */
-static char decompress_tmp_buff[64 * 1024];
-
 /* Used by some PostgreSQL macro definitions */
 #if PG_VERSION_NUM < 160000
 void
@@ -302,32 +293,18 @@ CopyAppend(const char *str)
 static int
 CopyAppendEncode(const char *str, int orig_len)
 {
-	/*
-	 * Should be enough in most cases. If it's not user can manually change
-	 * this limit. Unfortunately there is no way to know how much memory user
-	 * is willing to allocate.
-	 */
-	static char tmp_buff[64 * 1024];
-
-	/* Reserve one byte for a trailing zero. */
-	const int	max_offset = sizeof(tmp_buff) - 2;
 	int			curr_offset = 0;
 	int			len = orig_len;
+	char	   *tmp_buff = malloc(2 * orig_len + 1);
+
+	if (tmp_buff == NULL)
+	{
+		perror("malloc");
+		exit(1);
+	}
 
 	while (len > 0)
 	{
-		/*
-		 * Make sure there is enough free space for at least one special
-		 * symbol and a trailing zero.
-		 */
-		if (curr_offset > max_offset - 2)
-		{
-			printf("ERROR: Unable to properly encode a string since it's too "
-				   "large (%d bytes). Try to increase tmp_buff size in CopyAppendEncode "
-				   "procedure.\n", orig_len);
-			exit(1);
-		}
-
 		/*
 		 * Since we are working with potentially corrupted data we can
 		 * encounter \0 as well.
@@ -375,6 +352,8 @@ CopyAppendEncode(const char *str, int orig_len)
 
 	tmp_buff[curr_offset] = '\0';
 	CopyAppend(tmp_buff);
+	free(tmp_buff);
+
 	return 0;
 }
 
@@ -391,136 +370,151 @@ CopyAppendEncode(const char *str, int orig_len)
 static int
 CopyAppendNumeric(const char *buffer, int num_size)
 {
-       struct NumericData num;
+	struct NumericData *num = (struct NumericData *) malloc(num_size);
 
-       num = *(struct NumericData *)buffer;
-       if (NUMERIC_IS_SPECIAL(&num))
-       {
-               if (NUMERIC_IS_NINF(&num))
-               {
-                       CopyAppend("-Infinity");
-                       return 0;
-               }
-               if (NUMERIC_IS_PINF(&num))
-               {
-                       CopyAppend("Infinity");
-                       return 0;
-               }
-               if (NUMERIC_IS_NAN(&num))
-               {
-                       CopyAppend("NaN");
-                       return 0;
-               }
-               return -2;
-       }
-       else
-       {
-               int				sign;
-               int				weight;
-               int				dscale;
-               int				ndigits;
-               int				i;
-               char			   *str;
-               char			   *cp;
-               char			   *endcp;
-               int				d;
-               bool				putit;
-               NumericDigit		d1;
-               NumericDigit		dig;
-               NumericDigit	   *digits;
+	if (num == NULL)
+		return -2;
 
-               sign = NUMERIC_SIGN(&num);
-               weight = NUMERIC_WEIGHT(&num);
-               dscale = NUMERIC_DSCALE(&num);
+	memcpy((char *) num, buffer, num_size);
 
-               if (num_size == NUMERIC_HEADER_SIZE(&num))
-               {
-                       /* No digits - compressed zero. */
-                       CopyAppendFmt("%d", 0);
-                       return 0;
-               }
-               else
-               {
-                       ndigits = num_size / sizeof(NumericDigit);
-                       digits = (NumericDigit *)(buffer + NUMERIC_HEADER_SIZE(&num));
-                       i = (weight + 1) * DEC_DIGITS;
-                       if (i <= 0)
-                               i = 1;
+	if (NUMERIC_IS_SPECIAL(num))
+	{
+		int	result = -2;
 
-                       str = palloc(i + dscale + DEC_DIGITS + 2);
-                       cp = str;
+		if (NUMERIC_IS_NINF(num))
+		{
+			CopyAppend("-Infinity");
+			result = 0;
+		}
+		if (NUMERIC_IS_PINF(num))
+		{
+			CopyAppend("Infinity");
+			result = 0;
+		}
+		if (NUMERIC_IS_NAN(num))
+		{
+			CopyAppend("NaN");
+			result = 0;
+		}
 
-                       /*
-                        * Output a dash for negative values
-                        */
-                       if (sign == NUMERIC_NEG)
-                               *cp++ = '-';
+		free(num);
 
-                       /*
-                        * Output all digits before the decimal point
-                        */
-                       if (weight < 0)
-                       {
-                               d = weight + 1;
-                               *cp++ = '0';
-                       }
-                       else
-                       {
-                               for (d = 0; d <= weight; d++)
-                               {
-                                       dig = (d < ndigits) ? digits[d] : 0;
-                                       /* In the first digit, suppress extra leading decimal zeroes */
-                                       putit = (d > 0);
-                                               d1 = dig / 1000;
-                                       dig -= d1 * 1000;
-                                       putit |= (d1 > 0);
-                                       if (putit)
-                                               *cp++ = d1 + '0';
-                                       d1 = dig / 100;
-                                       dig -= d1 * 100;
-                                       putit |= (d1 > 0);
-                                       if (putit)
-                                               *cp++ = d1 + '0';
-                                       d1 = dig / 10;
-                                       dig -= d1 * 10;
-                                       putit |= (d1 > 0);
-                                       if (putit)
-                                               *cp++ = d1 + '0';
-                                       *cp++ = dig + '0';
-                               }
-                       }
+		return result;
+	}
+	else
+	{
+		int			sign;
+		int			weight;
+		int			dscale;
+		int			ndigits;
+		int			i;
+		char	   *str;
+		char	   *cp;
+		char	   *endcp;
+		int			d;
+		bool		putit;
+		NumericDigit d1;
+		NumericDigit dig;
+		NumericDigit *digits;
 
-                       /*
-                        * If requested, output a decimal point and all the digits that follow it.
-                        * We initially put out a multiple of DEC_DIGITS digits, then truncate if
-                        * needed.
-                        */
-                       if (dscale > 0)
-                       {
-                               *cp++ = '.';
-                               endcp = cp + dscale;
-                               for (i = 0; i < dscale; d++, i += DEC_DIGITS)
-                               {
-                                       dig = (d >= 0 && d < ndigits) ? digits[d] : 0;
-                                       d1 = dig / 1000;
-                                       dig -= d1 * 1000;
-                                       *cp++ = d1 + '0';
-                                       d1 = dig / 100;
-                                       dig -= d1 * 100;
-                                       *cp++ = d1 + '0';
-                                       d1 = dig / 10;
-                                       dig -= d1 * 10;
-                                       *cp++ = d1 + '0';
-                                       *cp++ = dig + '0';
-                               }
-                               cp = endcp;
-                       }
-                       *cp = '\0';
-                       CopyAppend(str);
-                       pfree(str);
-                       return 0;
-               }
-       }
+		sign = NUMERIC_SIGN(num);
+		weight = NUMERIC_WEIGHT(num);
+		dscale = NUMERIC_DSCALE(num);
+
+		if (num_size == NUMERIC_HEADER_SIZE(num))
+		{
+			/* No digits - compressed zero. */
+			CopyAppendFmt("%d", 0);
+			free(num);
+			return 0;
+		}
+		else
+		{
+			ndigits = num_size / sizeof(NumericDigit);
+			digits = (NumericDigit *) ((char *) num + NUMERIC_HEADER_SIZE(num));
+			i = (weight + 1) * DEC_DIGITS;
+			if (i <= 0)
+				i = 1;
+
+			str = palloc(i + dscale + DEC_DIGITS + 2);
+			cp = str;
+
+			/*
+			 * Output a dash for negative values
+			 */
+			if (sign == NUMERIC_NEG)
+				*cp++ = '-';
+
+			/*
+			 * Output all digits before the decimal point
+			 */
+			if (weight < 0)
+			{
+				d = weight + 1;
+				*cp++ = '0';
+			}
+			else
+			{
+				for (d = 0; d <= weight; d++)
+				{
+					dig = (d < ndigits) ? digits[d] : 0;
+
+					/*
+					 * In the first digit, suppress extra leading decimal
+					 * zeroes
+					 */
+					putit = (d > 0);
+					d1 = dig / 1000;
+					dig -= d1 * 1000;
+					putit |= (d1 > 0);
+					if (putit)
+						*cp++ = d1 + '0';
+					d1 = dig / 100;
+					dig -= d1 * 100;
+					putit |= (d1 > 0);
+					if (putit)
+						*cp++ = d1 + '0';
+					d1 = dig / 10;
+					dig -= d1 * 10;
+					putit |= (d1 > 0);
+					if (putit)
+						*cp++ = d1 + '0';
+					*cp++ = dig + '0';
+				}
+			}
+
+			/*
+			 * If requested, output a decimal point and all the digits that
+			 * follow it. We initially put out a multiple of DEC_DIGITS
+			 * digits, then truncate if needed.
+			 */
+			if (dscale > 0)
+			{
+				*cp++ = '.';
+				endcp = cp + dscale;
+				for (i = 0; i < dscale; d++, i += DEC_DIGITS)
+				{
+					dig = (d >= 0 && d < ndigits) ? digits[d] : 0;
+					d1 = dig / 1000;
+					dig -= d1 * 1000;
+					*cp++ = d1 + '0';
+					d1 = dig / 100;
+					dig -= d1 * 100;
+					*cp++ = d1 + '0';
+					d1 = dig / 10;
+					dig -= d1 * 10;
+					*cp++ = d1 + '0';
+					*cp++ = dig + '0';
+				}
+				cp = endcp;
+			}
+			*cp = '\0';
+			CopyAppend(str);
+			pfree(str);
+			free(num);
+			return 0;
+		}
+	}
 }
 
 /* Discard accumulated COPY line */
@@ -1108,10 +1102,35 @@ extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size,
 		{
 			result = ReadStringFromToast(buffer, buff_size, out_size, parse_value);
 		}
-		else
+		else if (VARATT_IS_EXTERNAL_ONDISK(buffer))
 		{
-			CopyAppend("(TOASTED)");
+			varatt_external toast_ptr;
+			VARATT_EXTERNAL_GET_POINTER(toast_ptr, buffer);
+			if (VARATT_EXTERNAL_IS_COMPRESSED(toast_ptr))
+			{
+#if PG_VERSION_NUM >= 140000
+				switch (VARATT_EXTERNAL_GET_COMPRESS_METHOD(toast_ptr))
+				{
+					case TOAST_PGLZ_COMPRESSION_ID:
+#endif
+						CopyAppend("(TOASTED,pglz)");
+#if PG_VERSION_NUM >= 140000
+						break;
+					case TOAST_LZ4_COMPRESSION_ID:
+						CopyAppend("(TOASTED,lz4)");
+						break;
+					default:
+						CopyAppend("(TOASTED,unknown)");
+						break;
+				}
+#endif
+			}
+			else
+				CopyAppend("(TOASTED,uncompressed)");
 		}
+		/* If tag is indirect or expanded, it was stored in memory. */
+		else
+			CopyAppend("(TOASTED IN MEMORY)");
 
 		*out_size = padding + len;
 		return result;
@@ -1156,6 +1175,7 @@ extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size,
 		int						decompress_ret;
 		uint32					len = VARSIZE_4B(buffer);
 		uint32					decompressed_len = 0;
+		char				   *decompress_tmp_buff;
 #if PG_VERSION_NUM >= 140000
 		ToastCompressionId		cmid;
 #endif
@@ -1166,19 +1186,13 @@ extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size,
 		decompressed_len = VARRAWSIZE_4B_C(buffer);
 #endif
 
-
 		if (len > buff_size)
 			return -1;
 
-		if (decompressed_len > sizeof(decompress_tmp_buff))
+		if ((decompress_tmp_buff = malloc(decompressed_len)) == NULL)
 		{
-			printf("WARNING: Unable to decompress a string since it's too "
-				   "large (%d bytes after decompressing). Consider increasing "
-				   "decompress_tmp_buff size.\n", decompressed_len);
-
-			CopyAppend("(COMPRESSED)");
-			*out_size = padding + len;
-			return 0;
+			perror("malloc");
+			exit(1);
 		}
 
 #if PG_VERSION_NUM >= 140000
@@ -1210,14 +1224,16 @@ extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size,
 
 		if ((decompress_ret != decompressed_len) || (decompress_ret < 0))
 		{
-			printf("WARNING: Unable to decompress a string. Data is corrupted.\n");
-			CopyAppend("(COMPRESSED)");
+			printf("WARNING: Corrupted toast data, unable to decompress.\n");
+			CopyAppend("(inline compressed, corrupted)");
 			*out_size = padding + len;
+			free(decompress_tmp_buff);
 			return 0;
 		}
 
 		result = parse_value(decompress_tmp_buff, decompressed_len);
 		*out_size = padding + len;
+		free(decompress_tmp_buff);
 		return result;
 	}
 
@@ -1366,23 +1382,22 @@ ReadStringFromToast(const char *buffer,
 #else
 		toast_ext_size = toast_ptr.va_extsize;
 #endif
+		num_chunks = (toast_ext_size - 1) / TOAST_MAX_CHUNK_SIZE + 1;
 
 		printf("  TOAST value. Raw size: %8d, external size: %8d, "
-				"value id: %6d, toast relation id: %6d\n",
+				"value id: %6d, toast relation id: %6d, chunks: %6d\n",
 				toast_ptr.va_rawsize,
 				toast_ext_size,
 				toast_ptr.va_valueid,
-				toast_ptr.va_toastrelid);
-
-		num_chunks = (toast_ext_size - 1) / TOAST_MAX_CHUNK_SIZE + 1;
-		printf("  Number of chunks: %d\n", num_chunks);
+				toast_ptr.va_toastrelid,
+				num_chunks);
 
 		/* Open TOAST relation file */
 		toast_relation_path = strdup(fileName);
 		get_parent_directory(toast_relation_path);
-		sprintf(toast_relation_filename, "%s/%d", toast_relation_path,
+		sprintf(toast_relation_filename, "%s/%d",
+				*toast_relation_path ? toast_relation_path : ".",
 				toast_ptr.va_toastrelid);
-		printf("  Read TOAST relation %s\n", toast_relation_filename);
 		toast_rel_fp = fopen(toast_relation_filename, "rb");
 		if (!toast_rel_fp) {
 			printf("Cannot open TOAST relation %s\n",
